program pascalmintv14;


{$APPTYPE CONSOLE}

uses
  SysUtils;

(* interpreter for Pascal-M  M-code

   H.J.C. Otten

   Accepts object-files produced with Pascal-M compiler of
   same version.
   Loads and executes program, debug function supported  *)

(* Version 1.4  november 2003, 2020

   Last update 13 aug 2020 Hans Otten  *)

(* Data-representation :

     character : 1 byte on stack

        +------------------+
Addr -> ! ascii code char  !
        +------------------+

     boolean : 2 bytes on stack

        +------------------+
        !         0        !
        +------------------+
Addr -> !  true or false   ! true = 1, false = 0
        +------------------+

     integer : 2 bytes on stack

        +-------------------+
        ! most sign part    ! div 256  signed
        +-------------------+
Addr -> ! least sign part   ! mod 256  signed
        +-------------------+

     pointer : 2 bytes on stack

        +-------------------+
        ! most sign part    ! div 256  signed
        +-------------------+
Addr -> ! least sign part   ! mod 256  signed
        +-------------------+

   Addresses, pointers such as stackpointer, markpointer etc
   always point to the least significant byte of the data-item.
   Examples are give above.

   Stack layout as given in the interpreting procedures
   gives an impression before and after the instruction
   execution. The stack is actually growing from high addresses
   to low addresses. For example, reserving space on the
   stack requires decrementing of the stackpointer.


   Memory layout :

   addresses

       0    +----------------------------+
            !                            !
            !                            !
                  code from compiler
            !                            !
            !                            !
            +----------------------------+
            !                            !
            !                            !
                  heap space
            !                            !
            !                            !
            +----------------------------+
            !                            !
            !                            !
                  free space > 40
            !                            !
            !                            !
            +----------------------------+
            !                            !
            !                            !
                  stack space
            !                            !
            !                            !
            +----------------------------+
            !     auto start code        !
            !                            !
   maxstore +----------------------------+
                                                   *)


(************************************************************)
(*							    *)
(*    Herafter follows the declarations for the interpreter *)
(*							    *)
(************************************************************)


CONST

  maxstore    = 2000 ;
  maxline     = 132 ;
  maxchcnt    = 132 ;
  maxproc     = 100 ;
  alphalength = 80 ;
  shlen       = 8 ;
  stackmin    = 40 ;
  CR          = 13 ; (* character value of Carriage Return *)
  LF          = 10 ; (* character value of LineFeed *)
  CTRLZ       = 26 ; (* Control-Z for end of line on INPUT *)
  maxbreak    =  8 ; (* maximum number of breakpoints *)
  maxfiles    =  7 ; (* maximum number of files including input and output *)

TYPE

  strng = packed array [1..maxchcnt] of char ;
  filestatus = ( correct, endoffile, overflowbuffer, checkerror ) ;
  alpha = packed array [ 1 .. alphalength ] of char ;

  procinfo = RECORD
               procaddress : integer ;
               procname    : alpha   ;
              END ;

  string20 = packed array [1..20] of char ;
  statusfile = (opened, fatalerror, closed) ;
  inOrOut    = ( inp, outp) ;
  fileptr    = ^fileinfo ;
  fileinfo   = RECORD
                 filename : text ;
                 name : string20 ;
                 InOut : inOrOut ;
                 linepos : integer ;
                 line : strng;
                 lengthline : integer ;
                 filestatus : statusfile ;
                 endoffile, endofline : boolean ;
               END ;

VAR

  filenumber : integer ;

  (* file variables *)

  pfiles : array [ 1..maxfiles] of fileptr ;
  kar : char ;
  linelength, recnr    : integer ;
  checksum, rchecksum  : integer ;
  fstatus              : filestatus ;
  line  : array [ 1 .. maxline ] of char ;
  objectfile : text ;

  (* procedure information buffer *)

  procbuffer : array [ 0 .. maxproc ] of procinfo ;
  procf : integer ; (* total number of proc found *)

  (* memory *)

  pc : integer ;
  stack, heap   : integer ;
  opcode, level : integer ;
  markp, markpsave : integer ;
  ldaddr : integer ;
  procaddr : integer ;
  store : array [ 0 .. maxstore ] of integer ;

  (* general *)

  i : integer ;
  loaderror, debug, continue, break : boolean ;
  breaktable : array [1..maxbreak] of integer ;


(************************************************************)
(*							                                            *)
(*    Herafter follows the code for operating system        *)
(*    file-handling support for the M-code instructions     *)
(*							                                            *)
(************************************************************)

  PROCEDURE FatalFileError ( errornr : integer ; errorfile : fileptr ) ;
  (********************************************************************)

    (* reports a fatal error occurred in file-handling the
       interpreter instructions and stops execution         *)

    BEGIN

      writeln ;
      writeln ('Program terminated, fatal file processing error') ;
      write ( 'File = ' , errorfile^.name, ' reason =' ) ;
      CASE errornr of
        1 : write (' reading character, not opened for input or fatal error' ) ;
        2 : write (' reading line, not opened for input or fatal error' ) ;
        3 : write (' writing line, not opened for output or fatal error ' ) ;
        4 : write (' writing character, not opened for output or fatal error') ;
        5 : write (' not opened or fatal error' ) ;
        6 : write (' output line exceeded ' ) ;
        7 : write (' not input file' )
      END ;
      writeln ;
      halt ;

    END ; (* FatalFileError *)

  PROCEDURE GetFileLine ( var infile : fileptr ) ;
  (**********************************************)

    (* gets a line from an input-file, called by
       openfile and GetFileChar.
       Returns contents in Line and nr of char's
       in linelength                             *)

   (* reads line from file into line *)

    BEGIN

       WITH infile^ do
         BEGIN
                 linepos := 0 ;
                 lengthline  := 1 ;
                 WHILE ( NOT eoln(filename) ) AND
                    ( lengthline < maxchcnt ) DO
                   BEGIN
                     read ( filename, line [ lengthline ] )  ;
                     lengthline := lengthline + 1 ;
                   END
         END

    END ; (* GetFileLine *)


  PROCEDURE Openfile ( var openfile  : fileptr ) ;
  (**********************************************)

    BEGIN

      WITH openfile^ DO
        BEGIN
          IF filestatus = opened
            THEN
              close (filename) ;
          IF InOut = inp
            THEN
              assign ( filename,
                     name)
            ELSE
              assign( filename,
                      name) ;

                IF InOut = inp
                  THEN
                    BEGIN
                      endoffile := false ;
                      endofline := false ;
                      reset (filename) ;
                      GetFileLine ( openfile)
                    END
                  ELSE
                    BEGIN
                      rewrite (filename) ;
                      linepos := 0
                    END ;
                filestatus := opened ;

        END ;

    END ; (* Openfile *)

  PROCEDURE Closefile ( VAR filetoclose  : fileptr ) ;
  (**************************************************)

    BEGIN

      WITH filetoclose^ DO
        BEGIN
          IF filenumber > 2
            THEN
              IF filestatus = opened
                THEN
                  BEGIN
                    (* flush output buffer if necessary *)
                    IF InOut = outp
                      THEN
                        IF linepos > 0
                          THEN
                            BEGIN
                              FOR i := 1 TO linepos DO
                                write ( filename, line[i] );
                              writeln ( filename )
                            END ;
                    close (filename) ;
                    filestatus := closed
                  END
        END

    END ; (* Closefile *)

  FUNCTION GetFileChar ( var infile : fileptr ) : char ;
  (****************************************************)

  (* reads buffered character from infile *)

  BEGIN

    WITH infile^ DO
      BEGIN
        IF filestatus <> opened
          THEN
            FatalFileError ( 5 , infile ) ;
        IF InOut = outp
          THEN
            FatalFileError ( 1 , infile  ) ;
        IF endoffile
          THEN
             BEGIN
               GetFileChar := ' ' ;
               endofline := true
             END
          ELSE
            BEGIN
              IF endofline
                THEN
                  BEGIN
                    endofline := false ;
                    readln ( filename ) ;
                    GetFileLine (infile) ;
                    GetFileChar := ' '
                  END
                ELSE
                  BEGIN
                    linepos := linepos + 1 ;
                    GetFileChar := line[linepos ] ;
                    IF (linepos + 1) >= lengthline
                      THEN
                        endofline := true
                  END
              END
          END

  END ; (* GetFileChar *)

  PROCEDURE GetLineFile ( var infile : fileptr ) ;
  (**********************************************)

  (* skip end of line in input file *)

  BEGIN

    WITH infile^ DO
      BEGIN
        IF filestatus <> opened
          THEN
            FatalFileError ( 5 , infile) ;
        IF InOut = outp
          THEN
            FatalFileError ( 2 , infile ) ;
        readln ( filename ) ;
        GetFileLine ( infile ) ;
        endofline := (lengthline = 1)
      END

  END ; (* GetLineFile*)

  PROCEDURE PutLineFile ( var outfile : fileptr ) ;
  (***********************************************)

  VAR

    i : integer ;

  BEGIN

    WITH outfile^ DO
      BEGIN
        IF filestatus <> opened
          THEN
            FatalFileError ( 5 , outfile ) ;
        IF InOut = inp
          THEN
            FatalFileError ( 3 , outfile) ;
        FOR i := 1 TO linepos DO
          write ( filename, line[i] );
        writeln ( filename ) ;
        linepos := 0 ;
      END

  END ; (* PutLineFile *)

  PROCEDURE PutFileChar ( outfile : fileptr ; kar : char ) ;
  (********************************************************)

  BEGIN

    WITH outfile^ DO
      BEGIN
        IF filestatus <> opened
          THEN
            FatalFileError ( 5, outfile ) ;
        IF InOut = inp
          THEN
            FatalFileError ( 4 , outfile) ;
        linepos := linepos + 1 ;
        line[linepos] := kar ;
        IF linepos >= maxchcnt
          THEN
            FatalFileError ( 6 , outfile )
      END

  END ; (* PutFileChar *)

PROCEDURE PutOutputChar ( kar : char ) ;
(**************************************)

  BEGIN

   write(kar) ;

  END ; (* PutOutputChar *)

FUNCTION GetKeyChar : char ;

  VAR

    kar     : char ;

  BEGIN

    read(kar) ;
    GetKeyChar := kar ;

  END ; (* GetKeyChar *)

FUNCTION GetInputChar : char ;
(****************************)

  VAR

    kar     : char ;

  BEGIN


     Read(kar) ;
    (* echo linefeed if carriage return entered *)

    GetInputChar := kar

  END ; (* GetInputChar *)


  PROCEDURE PutChar ( kar : char ) ;
  (********************************)

  (* Puts the character kar on the output-device
     pointed at by filenumber.
     Returns in status the resulting status. *)

  BEGIN

    IF filenumber = 2
      THEN
        PutOutputChar (kar)
      ELSE
        PutFileChar ( pfiles[filenumber],kar)

  END ; (* PutChar *)

  FUNCTION GetChar : char  ;
  (************************)

  (* Gets the character from the input-device
     pointed at by filenumber.                     *)

  BEGIN

    IF filenumber = 1
      THEN
        Getchar := GetInputChar
    ELSE IF filenumber = 3
      THEN
        GetChar := GetKeyChar
    ELSE
      GetChar := GetFileChar ( pfiles[filenumber] ) ;

  END ; (* GetChar *)


  PROCEDURE FileInitialize ;
  (************************)

    VAR

      i : integer ;

    BEGIN

      FOR i := 1 to maxfiles DO
        BEGIN
          new (pfiles[i]) ;
          pfiles[i]^.filestatus := closed
        END ;

      (* special actions for files INPUT and OUTPUT *)

      WITH pfiles[1]^ DO
        BEGIN
          filestatus := opened ;
          endoffile := false ;
          endofline := false ;
          name := 'INPUT               '
        END ;
      WITH pfiles[3]^ DO
        BEGIN
          filestatus := opened ;
          endoffile := false ;
          endofline := false ;
          name := 'KEYBOARD            '
        END ;
      pfiles[2]^.filestatus := opened ;
      pfiles[2]^.name := 'OUTPUT              '

    END ; (* FileInitialize *)


(************************************************************)
(*							                                            *)
(*    Herafter follows the code to load the M-code program  *)
(*							                                            *)
(************************************************************)


  PROCEDURE GetLine ( VAR linelength : integer ;
  (***************)   VAR fstatus : filestatus ) ;

    BEGIN

     if eof ( objectfile )
       then
         fstatus := endoffile
       ELSE
         BEGIN
           recnr := recnr + 1 ;
           linelength := 1 ;
           read (objectfile, line [ linelength ] ) ;
           while not eoln(objectfile ) and ( linelength < maxline ) do
             BEGIN
               linelength := linelength + 1 ;
               read ( objectfile, line [ linelength ] )  ;
             END ;
           if not eoln (objectfile )
             then
               fstatus := overflowbuffer ;
           readln (objectfile ) ;
         END ;

    END ; (* GetLine *)

  FUNCTION PackHex ( startposition : integer ) : integer ;
  (******************************************************)

  VAR

    hexnumber, chvalue, i, j  : integer ;

  BEGIN

   hexnumber := 0 ;
   FOR i := 0 TO 1 DO
     BEGIN
       chvalue := ord ( line[startposition + i ] ) ;
       IF ( chvalue >= ord ('0') ) AND
          ( chvalue <= ord ('9') )
         THEN
           j := chvalue - ord ('0')
         ELSE
           j := chvalue - ord ('A') +  10 ;
       hexnumber := hexnumber * 16 + j
      END ;
    PackHex := hexnumber ;

  END ; (* PackHex *)

PROCEDURE Loadp1 ;
 (****************)

 (* Loads P1 record into store *)

 VAR

   nrLoadbytes,i : integer ;

 BEGIN

   nrLoadbytes := PackHex ( 3 ) ;
   checksum := nrLoadbytes ;
   for i := 1 to nrLoadbytes - 1 do
     BEGIN
       store[pc] := PackHex (  2*i - 1 + 4  ) ;
       checksum := checksum + store[pc] ;
       pc := pc + 1 ;
     END ;
   rchecksum := PackHex ( nrLoadbytes * 2 - 1 + 4 ) ;

   i := (16383 - checksum ) mod 256 ;

   IF i <> rchecksum
     THEN
       BEGIN
         writeln ('Checksum error in P1 record ') ;
         loaderror := true ;
         fstatus := checkerror
       END ;

   IF ( maxstore - pc ) < 40
     THEN
       BEGIN
         writeln ('Store exhausted in P1 record ') ;
         loaderror := true ;
         fstatus := checkerror
       END ;

 END ; (* Loadp1 *)

 PROCEDURE Loadp2 ;
  (*

  The P2 record is used for “satisfying” forward refrences by going back to a
  previously-loaded area in the program and inserting an address.
  In this case, the four hex digitis appearing after the P2 specify
  the byte address to be modified
  (relative to the beginning of the progam)
  and the next four hex digits specify the value to be inserted
  into that location and the following one
  (since the planted addresses are always two bytes in length).

   Loads P2 record with address into store buffer *)

  VAR

    loc, val, i, locl, loch, vall, valh  : integer ;

  BEGIN

    loch := PackHex(3) ;
    locl := PackHex(5) ;
    loc := 256* loch +  locl + ldaddr ;
    checksum := loch + locl ;
    valh := PackHex(7) ;
    vall := Packhex(9) ;
    val := 256* valh + vall ;
    checksum := checksum +  valh + vall ;
    store[loc    ] := val div 256  ;
    store[loc + 1] := val mod 256  ;
    rchecksum := PackHex ( 11 ) ;
    i := ( 16383 - checksum ) mod 256 ;
    if i <> rchecksum
      then
        BEGIN
          writeln ('Checksum error in P2 record ') ;
	  loaderror := true ;
          fstatus := checkerror
        END ;

  END ; (* Loadp2 *)

  PROCEDURE Loadp4 ;
  (****************)

  (* Loads P4 record into procedure buffer   *)
  (* line [3] .. line[4] is procedure number *)
  (* line [5] .. line [12] is procedure name *)

  VAR

    nrproc, i : integer ;

  BEGIN

    procf := procf + 1 ;
    nrproc := PackHex ( 3 ) ;
    checksum := nrproc ;
    for i := 1 to shlen do
     BEGIN
      procbuffer[nrproc].procname[i] := line [ 4 + i ] ;
      checksum := checksum + ord ( line[i+4] ) ;
     END ;
    procbuffer[nrproc].procaddress := pc  ;
    rchecksum := PackHex ( 13 ) ;

    i := ( 16383 - checksum ) mod 256 ;
    if i <> rchecksum
      then
        BEGIN
          writeln ('Checksum error in P4 record ') ;
	  loaderror := true ;
          fstatus := checkerror
        END ;

  END ; (* Loadp4 *)


  PROCEDURE Load ;
  (**************)

  (* Loads p-code output from compiler *)
  (* records output as followed by checksum
     P1 = code-bytes etc
     P2 = address-reference four bytes
     P4 = procedure address + procedure name
     P9 = END of code ( no checksum )       *)

  VAR

    i : integer ;

  BEGIN

    fstatus := correct ;
    recnr := 0 ;
    repeat
      GetLine ( linelength, fstatus ) ;
      checksum := 0 ;
      if fstatus <> endoffile
        then
          if line[1] = 'P'
            then
              case line[2] of
               '1' : Loadp1 ;
               '2' : Loadp2 ;
               '4' : Loadp4 ;
               '9' : fstatus := endoffile
              END ;

    until (fstatus = endoffile ) or ( fstatus = checkerror ) ;

    if fstatus = checkerror
      then
        writeln ('Loading ended due to checkerror in record : ',
                  recnr )
      ELSE if debug then
        BEGIN
          write ('Program ' ) ;
          for i := 1 to shlen do
                write(procbuffer[0].procname[i] ) ;
          writeln ( ' Loaded ' )
        END ;

  END ; (* Load *)

PROCEDURE WriteHex ( hexword : integer ) ;
(****************************************)

VAR

  highbyte, lowbyte : integer ;

BEGIN

  IF hexword < 0
    THEN
      BEGIN
        hexword := -hexword ;
        write ('-')
      END ;
  lowbyte  := hexword mod 16 ;
  highbyte := hexword div 16 ;
  if highbyte  < 10
    then
      write ( chr (ord ('0') + highbyte) )
    ELSE
      write ( chr (ord ('A') + highbyte - 10 ) ) ;
  if lowbyte < 10
    then
      write ( chr (ord ('0') + lowbyte) )
    ELSE
      write ( chr (ord ('A') + lowbyte   - 10 ) ) ;

END ; (* WriteHex *)


FUNCTION ReadHex ( Hexstrng: alpha ; Endline : integer ) : integer ;
(********************************************************************)

(* Returns from Hexstring an integer *)

VAR

  i,j,k : integer ;
  ch : char ;

BEGIN

  (* start at first position of strngand zero number *)
  j := 1 ;
  k := 0 ;

  (* skip leading blanks *)
  WHILE (hexstrng[j] = ' ') and ( j < ENDline )  DO
    j := j + 1 ;

  IF j < ENDline THEN
    BEGIN
      REPEAT
        ch := hexstrng[j] ;
        j := j + 1 ;
        IF ( ch >= '0' ) and ( ch <= '9' )
          THEN
            i := ord(ch) - ord('0')
        ELSE IF ( ch >= 'A') and ( ch <= 'F' )
          THEN
            i := ord(ch) - ord('A') + 10
        ELSE IF ( ch >= 'a') and ( ch <= 'f' )
          THEN
            i := ord(ch) - ord('a') + 10
        ELSE
          i := -1 ;
        IF i >= 0
          THEN
            k := 16*k + i
      UNTIL ( i < 0 ) or ( j > (Endline - 1) ) ;
    END ;
  ReadHex := k

END ; (* ReadHex *)



(************************************************************)
(*							    *)
(*    Instruction interpreting                              *)
(*    Hereafter follow the procedures for interpreting      *)
(*    the M-code instructions                               *)
(*							    *)
(************************************************************)

  PROCEDURE NotImplemented ;

    BEGIN

      writeln ;
      write ( 'Not implemented instruction ') ;
      writehex( store[ (pc - 1)] ) ;
      write (' at ') ;
      writehex ( (pc - 1) div 256 ) ;
      writehex ( (pc - 1) mod 256 ) ;
      writeln ;
      debug := true ;

     END ; (* NotImplemented *)

Function BaseAddress ( level : integer ) : integer ;
(**************************************************)

  (* returns base-address of stackframe of level
     by following baseaddress linkpointers :
     Note that all addresses are the address of the
     least significant byte of the item.

              +-----------------------+
      mp -->  ! link markpointer high ! --> previous frame
              +-----------------------+
              ! link markpointer low  !
              +-----------------------+   *)
 VAR

   temp : integer ;

 BEGIN

   temp := markp ;
   WHILE level > 0 DO
     BEGIN
       temp := store[temp] * 256 + store[temp - 1] ;
       level := level - 1
     END ;

    BaseAddress := temp + 1 ;

  END ;

PROCEDURE LDCIS ;
(***************)

  (* pushes 'level' as small constant on stack *)

  BEGIN

    store[stack    ] := 0     ;   (* high byte zero *)
    store[stack - 1] := level ;
    stack := stack - 2

  END ; (* LDCIS *)

PROCEDURE LDAS  ;
(***************)

  (* pushes address of variable at offset (following
     opcode) from base-address at level on stack,
     single byte offset < 256 *)

  VAR

    temp : integer ;

  BEGIN

    temp := BaseAddress( level ) - store[ pc ] ;

    store[stack    ] := temp div 256 ; (* high byte of address *)
    store[stack - 1] := temp mod 256 ; (*  low byte of address *)
    stack := stack - 2 ;

    pc := pc + 1

  END ; (* LDAS *)

PROCEDURE LDA   ;
(***************)

  (* pushes address of variable at offset (following
     opcode) from base-address at level on stack,
     double byte offset, high byte first after PC *)

  VAR

    temp : integer ;

  BEGIN

    temp := BaseAddress( level ) - store[ pc ] * 256 - store [pc + 1] ;

    store[stack    ] := temp div 256 ; (* high byte of address *)
    store[stack - 1] := temp mod 256 ; (*  low byte of address *)
    stack := stack - 2 ;

    pc := pc + 2

  END ; (* LDA *)

PROCEDURE MST0  ;
(***************)

  (* marks stack without return bytes, sets up
     stack frame with room for : base-address link
                                 markpointer save
                                 programcounter  *)

  VAR

     temp : integer ;

  BEGIN

    temp := BaseAddress(level) - 1 ;
    markpsave := stack ;
    store [stack    ] := temp div 256 ;
    store [stack - 1] := temp mod 256 ;
    store [stack - 2] := markp div 256 ;
    store [stack - 3] := markp mod 256 ;
    stack := stack - 6 ;

  END ;

PROCEDURE MSTN  ;
(***************)

  (* mark stack with room for returnvalue for
     function, nr of bytes found after opcode.
     Stack frame set up with MST0               *)

  BEGIN

    stack := stack - ( store [pc] ) ;
    MST0 ;
    pc := pc + 1

  END ;

PROCEDURE LOD1  ;
(***************)

  (* load 1 byte item on stack, at offset following opcode
     from base address at level, expanded to 16 bits *)

  BEGIN

    store [stack    ] := 0 ;
    store [stack - 1] := store [ BaseAddress (level) - store [pc] ] ;

    stack := stack - 2 ;
    pc := pc + 1

  END ; (* LOD1 *)

PROCEDURE LOD2  ;
(***************)

  (* load word item on stack, from offset following opcode
     from base address at level  *)
  VAR

    temp : integer ;

  BEGIN

    temp := BaseAddress (level) - store [pc] ;

    store [stack - 1] := store [ temp    ] ;
    store [stack    ] := store [ temp + 1] ;

    stack := stack - 2 ;
    pc := pc + 1

  END ; (* LOD2 *)

PROCEDURE STR1  ;
(***************)

  (* pulls word from stack and stores least significant
     byte at level at offset following opcode

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  value                     !<- stack
            +----------------------------+
    stack ->!                            !        *)



  BEGIN

    store [BaseAddress (level) - store [pc] ] := store[stack + 1] ;

    stack := stack + 2 ;
    pc := pc + 1

  END ;

PROCEDURE STR2  ;
(***************)

  (* pulls word from stack and stores word
     at level at offset following opcode

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  value                     !<- stack
            +----------------------------+
    stack ->!                            !        *)

  VAR

    temp : integer ;

  BEGIN

    temp := BaseAddress (level) - store [pc] ;

    store [temp    ] := store[stack + 1] ;
    store [temp + 1] := store[stack + 2] ;
    stack := stack + 2 ;
    pc := pc + 1

  END ; (* STR2 *)

PROCEDURE LEQ2  ;
(***************)

(* compares for less or equal
   condition tested : first value =< second value
   result : boolean 1 if less or equal else boolean 0

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first value / boolean     !
            +----------------------------+
            !  second value              !<- stack
            +----------------------------+
    stack ->!                            !        *)

  BEGIN

    IF ( ( store [stack + 3] +  256 * store [stack + 4] ) <=
         ( store [stack + 1] +  256 * store [stack + 2] )    )
      THEN
        store [stack + 3] := 1
      ELSE
        store [stack + 3] := 0 ;

    store [stack + 4] := 0 ;
    stack := stack + 2

  END ;  (* LEQ2 *)

PROCEDURE UJP ;
(***************)

  (* unconditional jump, new pc := value after opcode
     plus Load address *)

  BEGIN

    pc := store [pc] * 256 + store [pc + 1] + ldaddr;

  END ; (* UJP *)

PROCEDURE FJP   ;
(***************)

  (* false jump, taken if value on stack is false, zero
     jump performed with UJP

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  boolean                   ! <- stack
            +----------------------------+
    stack ->!                            !       *)

  BEGIN

    stack := stack + 2 ;
    IF store [ stack - 1] = 0
      THEN
        UJP
      ELSE
        pc := pc + 2

  END ; (* FJP *)

PROCEDURE MFOR   ;
(***************)

  (* implements the FOR statement.
     Controls the loop with information on the stack.
     Stored are
      - the address off the loop variable
      - the beginvalue
      - the endvalue
      - a flag word

     The FOR instruction implements a jump-instruction,
     the jump is taken as long as the end-condition is
     not reached.

     The stack is cleaned up if the end-condition is reached.

     The flag word has the following meaning :

     Bit 0 = 0 : the loop variable is a one byte location
     Bit 0 = 1 : the loop variable is a two byte location
     Bit 2 = 0 : to loop
     Bit 2 = 1 : downto loop
     Bit 8 = 0 : loop not initialized
     Bit 8 = 1 : loop initialized

     Stack layout :
     start                                  end

            !                            !<- stack if end
            +----------------------------+
            !  address of loop variable  !
            +----------------------------+
            !  beginvalue                !
            +----------------------------+
            !  endvalue                  !
            +----------------------------+
            !  flag word                 !
            +----------------------------+
    stack ->!                            !<- stack if not end *)

  VAR

    twobyte, down : boolean ;
    beginvalue, endvalue, curvalue, lpaddress : integer ;

  BEGIN

    down    :=  ( (store[stack + 1] div 4 ) = 1 ) ;
    twobyte :=  odd  (store[stack + 1] ) ;
    endvalue   := store[stack + 3] + 256 *  store[stack + 4] ;
    lpaddress  := store[stack + 7] + 256 *  store[stack + 8] ;

    (* loop initialized ? *)

    IF store[stack + 2] = 0
      THEN
        BEGIN
          beginvalue := store[stack + 5] + 256 *  store[stack + 6] ;
          IF down
            THEN
              beginvalue := beginvalue + 1
            ELSE
              beginvalue := beginvalue - 1 ;
          store[stack + 2] := 1 ;
          store[lpaddress] := beginvalue mod 256 ;
          IF twobyte
            THEN
              store[lpaddress + 1] := beginvalue div 256 ;
         END ;

    curvalue := store[lpaddress] ;
    IF twobyte
      THEN
        curvalue := curvalue + 256 * store[lpaddress + 1] ;

    IF curvalue = endvalue
      THEN
        BEGIN
          pc := pc + 2 ;
          stack := stack + 8
        END
      ELSE
        BEGIN
          IF down
            THEN
              curvalue := curvalue - 1
            ELSE
              curvalue := curvalue + 1 ;
          store[lpaddress] := curvalue mod 256 ;
          IF twobyte
            THEN
              store[lpaddress + 1] := curvalue div 256 ;
          UJP
        END

  END ; (* MFOR *)

PROCEDURE LEQM  ;
(***************)


(* compares for equality of arrays or records
   condition tested : first value <= second value
   result : boolean 1 if equal else boolean 0
   size of structure follows opcode

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first address / boolean   !
            +----------------------------+
            !  second address            !<--stack
            +----------------------------+
    stack ->!                            !        *)

  VAR

    size,count : integer ;
    first, second, a, b : integer ;
    lessorequal : boolean ;

  BEGIN

    size := store[pc] * 256 + store[pc + 1] ;
    first  := store[stack + 3] + 256 * store[stack + 4] ;
    second := store[stack + 1] + 256 * store[stack + 2] ;
    lessorequal := true ;
    count := 0 ;
    while (count < size) do
      BEGIN
        a := store [first  + count] ;
        b := store [second + count] ;
        IF a < b
          THEN
            count := size
        ELSE IF a = b
          THEN
            count := count + 1
        ELSE  (* a > b *)
          BEGIN
            lessorequal := false ;
            count := size
          END
      END ;
  if lessorequal
      then
        store[stack + 3] := 1
      else
        store[stack + 3] := 0 ;
    store [stack + 4] := 0 ;
    stack := stack + 2 ;
    pc := pc + 2 ;

  END ;  (* LEQM *)

PROCEDURE LES2  ;
(***************)


(* compares for less or equal
   condition tested : first value < second value
   result : boolean 1 if less or equal else boolean 0

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first value / boolean     !
            +----------------------------+
            !  second value              !<- stack
            +----------------------------+
    stack ->!                            !        *)

  BEGIN

    IF ( ( store [stack + 3] +  256 * store [stack + 4] ) <
         ( store [stack + 1] +  256 * store [stack + 2] )    )
      THEN
        store [stack + 3] := 1
      ELSE
        store [stack + 3] := 0 ;

    store [stack + 4] := 0 ;
    stack := stack + 2

  END ;  (* LEQ2 *)


PROCEDURE LEQ8  ;
(***************)

  (* checks if all elements of first set are present
     in second set. Leaves boolean result on the stack.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  first set / result boolean!
            +----------------------------+
            !  second set                ! <- stack
            +----------------------------+
    stack ->!                            !            *)

VAR

  firstvalue, secondvalue, resultvalue,
  bytecount, bitcount  : integer ;

  BEGIN

    resultvalue := 1 ;   (* start with true *)

    FOR bytecount := 1 to 8 DO
      BEGIN
        firstvalue  := store[stack + 8 + bytecount    ] ;
        secondvalue := store[stack + bytecount] ;
        FOR bitcount := 1 to 8 DO
          BEGIN
            IF ( odd(firstvalue) )
              THEN
                IF not (odd(secondvalue))
                  THEN
                    resultvalue := 0 ;
            firstvalue  := firstvalue  div 2 ;
            secondvalue := secondvalue div 2 ;
          END ;
      END;

    store[stack + 15] := resultvalue ;
    store[stack + 16] := 0 ;
    stack := stack + 14 ;

  END ; (* LEQ8 *)

PROCEDURE LESM  ;
(***************)


(* compares for equality of arrays or records
   condition tested : first value < second value
   result : boolean 1 if equal else boolean 0
   size of structure follows opcode

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first address / boolean   !
            +----------------------------+
            !  second address            !<--stack
            +----------------------------+
    stack ->!                            !        *)

  VAR

    size,count 		: integer ;
    first, second, a, b : integer ;
    less 		: boolean ;

  BEGIN

    size := store[pc] * 256 + store[pc + 1] ;
    first  := store[stack + 3] + 256 * store[stack + 4] ;
    second := store[stack + 1] + 256 * store[stack + 2] ;
    less := false ;
    count := 0 ;
    while (count < size) do
      BEGIN
        a := store [first  + count] ;
        b := store [second + count] ;
        IF a < b
          THEN
            BEGIN
              count := size ;
              less := true
            END
        ELSE IF a = b
          THEN
            count := count + 1
        ELSE  (* a > b *)
            count := size
      END ;
    IF less
      THEN
        store[stack + 3] := 1
      ELSE
        store[stack + 3] := 0 ;
    store [stack + 4] := 0 ;
    stack := stack + 2 ;
    pc := pc + 2 ;

  END ;  (* LESM *)

PROCEDURE EQU2  ;
(***************)

(* compares for equality
   condition tested : first value = second value
   result : boolean 1 if equal else boolean 0

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first value / boolean     !
            +----------------------------+
            !  second value              !<- stack
            +----------------------------+
    stack ->!                            !        *)


  BEGIN

    IF ( ( store [stack + 1] = store [stack + 3] ) AND
         ( store [stack + 2] = store [stack + 4] )      )
      THEN
        store [stack + 3] := 1
      ELSE
        store [stack + 3] := 0 ;

    store [stack + 4] := 0 ;
    stack := stack + 2

  END ;  (* EQU2 *)

PROCEDURE GEQ8  ;
(***************)

  (* checks if all elements of second set are present
     in first set. Leaves boolean result on the stack.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  first set / result boolean!
            +----------------------------+
            !  second set                ! <- stack
            +----------------------------+
    stack ->!                            !            *)

VAR

  firstvalue, secondvalue, resultvalue,
  bytecount, bitcount  : integer ;

  BEGIN

    resultvalue := 1 ;   (* start with true *)

    FOR bytecount := 1 to 8 DO
      BEGIN
        firstvalue  := store[stack + 8 + bytecount    ] ;
        secondvalue := store[stack + bytecount] ;
        FOR bitcount := 1 to 8 DO
          BEGIN
            IF ( odd(secondvalue) )
              THEN
                IF not (odd(firstvalue))
                  THEN
                    resultvalue := 0 ;
            firstvalue  := firstvalue  div 2 ;
            secondvalue := secondvalue div 2 ;
          END ;
      END;

    store[stack + 15] := resultvalue ;
    store[stack + 16] := 0 ;
    stack := stack + 14 ;

  END ; (* GEQ8 *)

PROCEDURE EQUM  ;
(***************)


(* compares for equality of arrays or records
   condition tested : first value = second value
   result : boolean 1 if equal else boolean 0
   size of structure follows opcode

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first address / boolean   !
            +----------------------------+
            !  second address            !<--stack
            +----------------------------+
    stack ->!                            !        *)

  VAR

    size,count : integer ;
    first, second : integer ;
    equal : boolean ;

  BEGIN

    size := store[pc] * 256 + store[pc + 1] ;
    first  := store[stack + 1] + 256 * store[stack + 2] ;
    second := store[stack + 3] + 256 * store[stack + 4] ;
    equal := true ;
    count := 0 ;
    while equal and (count < size) do
      begin
        equal := (store [first + count] = store [second + count ] ) ;
        count := count + 1
      end ;
  if equal
      then
        store[stack + 3] := 1
      else
        store[stack + 3] := 0 ;
    store [stack + 4] := 0 ;
    stack := stack + 2 ;
    pc := pc + 2 ;

  END ;  (* EQUM *)

PROCEDURE EQU8  ;
(***************)

(* compares for equality of sets
   condition tested : first value = second value
   result : boolean 1 if equal else boolean 0

     Stack layout :
     start                                  end

            !                            !
            +----------------------------+
            !  first set value / boolean !
            +----------------------------+
            !  second set value          !<--stack
            +----------------------------+
    stack ->!                            !        *)


  BEGIN

    IF ( ( store [stack + 1] = store [stack +  9] ) AND
         ( store [stack + 2] = store [stack + 10] ) AND
         ( store [stack + 3] = store [stack + 11] ) AND
         ( store [stack + 4] = store [stack + 12] ) AND
         ( store [stack + 5] = store [stack + 13] ) AND
         ( store [stack + 6] = store [stack + 14] ) AND
         ( store [stack + 7] = store [stack + 15] ) AND
         ( store [stack + 8] = store [stack + 16] )      )

      THEN
        store [stack + 15] := 1
      ELSE
        store [stack + 15] := 0 ;

    store [stack + 16] := 0 ;
    stack := stack + 14

  END ;  (* EQU8 *)


PROCEDURE IND1  ;
(***************)

  (* loads one byte value on stack from address on stack
     Stack layout :
     start                     end

            !                            !
            +----------------------------+
            !  address/value in low byte !
            +----------------------------+
    stack ->!                            !<- stack   *)

  VAR

    addr : integer ;

  BEGIN

    addr := store[stack + 1] + 256 * store[stack + 2] ;
    store[stack + 1] := store [addr] ;
    store[stack + 2] := 0 ;

  END ; (* IND1 *)

PROCEDURE IND2  ;
(***************)

  (* loads two byte value on stack from address on stack
     Stack layout :
     start                     end

            !                            !
            +----------------------------+
            !  address/value             !
            +----------------------------+
    stack ->!                            ! <- stack   *)

  VAR

    addr : integer ;

  BEGIN

    addr := store[stack + 1] + 256 * store[stack + 2] ;
    store[stack + 1] := store [addr    ] ;
    store[stack + 2] := store [addr + 1] ;

  END ; (* IND2 *)

PROCEDURE IND8  ;
(***************)


  (* loads two byte value on stack from address on stack
     Stack layout :
     start                     end

            !                            !
            +----------------------------+
            !  address/ set value        !
            +----------------------------+
    stack ->!  set value                 !
            +----------------------------+
            !  set value                 !
            +----------------------------+
            !  set value                 !
            +----------------------------+
            !                            !<- stack   *)

  VAR

    addr,i : integer ;

  BEGIN

    addr := store[stack + 1] + 256 * store[stack + 2] ;
    stack := stack - 6 ;
    FOR i := 0 TO 7 DO
      store[stack + 1 + i] := store [addr + i] ;

  END ; (* IND8 *)

PROCEDURE STO1  ;
(***************)
  (* indirectly stores 1 byte from stack at address on stack

     Stack layout :
     begin                                 end

            !                            ! <- stack
            +----------------------------+
            !              address       !
            +----------------------------+
            !          value low byte    !
            +----------------------------+
    stack ->!                            !                *)

  VAR

    addr : integer ;

  BEGIN

    addr := store[stack + 3] + 256 * store[stack + 4] ;
    store [addr] := store[stack + 1] ;
    stack := stack + 4 ;

  END ; (* STO1 *)

PROCEDURE ST02  ;
(***************)

  (* indirectly stores 2 bytes from stack at address on stack
     Stack layout :
     begin                                 end

            !                            ! <- stack
            +----------------------------+
            !         address            !
            +----------------------------+
            !         value              !
            +----------------------------+
    stack ->!                            !                *)


  VAR

    addr : integer ;

  BEGIN

    addr := store[stack + 3] + 256 * store[stack + 4] ;
    store [addr    ] := store[stack + 1] ;
    store [addr + 1] := store[stack + 2] ;
    stack := stack + 4 ;

  END ; (* STO2 *)


PROCEDURE STO8  ;
(***************)

  (* indirectly stores 2 bytes from stack at address on stack
     Stack layout :
     begin                                 end

            !                            ! <- stack
            +----------------------------+
            !         address            !
            +----------------------------+
            !         set value  8 bytes !
            +----------------------------+
    stack ->!                            !                *)


  VAR

    addr,i : integer ;

  BEGIN

    addr := store[stack + 9] + 256 * store[stack + 10] ;
    FOR i := 0 TO 7 DO
      store [addr + i] := store[stack + 1 + i] ;
    stack := stack + 10 ;

  END ; (* STO8 *)

PROCEDURE LDC   ;
(***************)

  (* loads constant following opcode on the stack
     Stack layout :
     begin                                 end
            !                            !
            +----------------------------+
    stack ->!  constant                  !
            +----------------------------+
            !                            !<- stack *)

  BEGIN

    stack := stack - 2 ;
    store [stack + 1] := store [pc + 1] ;
    store [stack + 2] := store [pc] ;

    pc := pc + 2

  END ;

PROCEDURE RETP  ;
(***************)

  (* Returns from procedure/function, cleans up
     stack by removing local variables and
     linkframe
     Situation before :

    before                     after
            +----------------------------+
  markp --> !          baseadr           ! <--- stack
            +----------------------------+
            !          mpsave            !
            +----------------------------+
            !          pcsave            !
            +----------------------------+
            !                            !
            !      local variables       !
            !                            !
            +----------------------------+
  stack --> !                            !             *)

  BEGIN

    stack := markp ;
    markp := store [stack - 2] * 256 + store [stack - 3] ;
    pc    := store [stack - 4] * 256 + store [stack - 5] ;

  END ;

PROCEDURE ADI   ;
(***************)

  (* adds two integers on stack, leaves integer
     result on stack.
     evaluates expression :
         result := first value / second value

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  value / result     !
            +----------------------------+
            !  second value              ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  VAR

    result : integer ;

  BEGIN

    result := ( store [stack + 1] + 256 * store [stack + 2] ) +
              ( store [stack + 3] + 256 * store [stack + 4] ) ;
    store [stack + 3] := result mod 256 ;
    store [stack + 4] := result div 256 ;
    stack := stack + 2 ;

  END ; (* ADI *)

PROCEDURE MAND   ;
(***************)

  (* Performs AND on two booleans on the stack,
     leaves boolean result on stack.
     Booleans placed on stack as two bytes,
     low byte contains 0 if false.
     evaluates expression :
     result := first boolean AND  second boolean.
     (true is 1 if first and second true)

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  boolean/ result    !
            +----------------------------+
            !  second boolean            ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  BEGIN

    IF (store[stack + 1] = 1) AND  (store[stack + 3] = 1)
      THEN
        store[stack + 3] := 1
      ELSE
        store[stack + 3] := 0 ;
    stack := stack + 2 ;

  END ; (* MAND *)


PROCEDURE DIF   ;
(***************)

  (* gives the differnce of two sets on the stack,
     leaves result on the stack. Collects in result
     all elements present in first set and not in second set.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  first set / result set    !
            +----------------------------+
            !  second set                ! <- stack
            +----------------------------+
    stack ->!                            !            *)

VAR

  firstvalue, secondvalue, resultvalue,
  k, bytecount, bitcount  : integer ;

  BEGIN

    FOR bytecount := 1 to 8 DO
      BEGIN
        resultvalue := 0 ;
        k := 1 ;
        firstvalue  := store[stack + 8 + bytecount    ] ;
        secondvalue := store[stack + bytecount] ;
        FOR bitcount := 1 to 8 DO
          BEGIN
            IF ( odd(firstvalue) )  and ( not (odd(secondvalue)) )
              THEN
                resultvalue := resultvalue + k ;
            k := k * 2 ;
            firstvalue  := firstvalue  div 2 ;
            secondvalue := secondvalue div 2
          END ;
        store[stack + 8 + bytecount] := resultvalue ;
      END;
    stack := stack + 8 ;

  END ; (* DIF *)

PROCEDURE DVI   ;
(***************)

  (* divides two integers on stack, leaves integer
     result on stack. Signed division.
     evaluates expression :
         result := first value / second value

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  value / result     !
            +----------------------------+
            !  second value              ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  VAR

    result : integer ;

  BEGIN

    result := ( store[stack + 3] + 256 * store[stack + 4] ) div
              ( store[stack + 1] + 256 * store[stack + 2] )  ;
    store[stack + 3] := result mod 256 ;
    store[stack + 4] := result div 256 ;
    stack := stack + 2

  END ; (* DVI *)

PROCEDURE INN   ;
(***************)

  (* The is member of test for sets is implemented
     with this instruction. Leaves boolean true
     on stack if member in set.

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  number of member/ result  !
            +----------------------------+
            !  set value                 ! <- stack
            +----------------------------+
    stack ->!                            !           *)


  VAR

    bytenumber, bitnumber, setvalue : integer ;

  BEGIN


    bytenumber := store[stack + 9] ;
    bitnumber := bytenumber mod 8 ;
    bytenumber := bytenumber div 8 ;

    setvalue := store[stack + 1 + bytenumber] ;
    while (7 - bitnumber) > 0 do
      begin
        setvalue := setvalue div 2 ;
        bitnumber := bitnumber + 1
      end ;
    if odd (setvalue)
      then
        store[stack + 9] := 1
      else
        store[stack + 9] := 0 ;
    store[stack + 10] := 0 ;
    stack := stack + 8 ;

  END ; (* INN *)

PROCEDURE INT   ;
(***************)

  (* gives the intersection of two sets on the stack,
     leaves result on the stack. Collects in result
     all elements present in first and in second set.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  first set / result set    !
            +----------------------------+
            !  second set                ! <- stack
            +----------------------------+
    stack ->!                            !            *)

VAR

  firstvalue, secondvalue, resultvalue,
  k, bytecount, bitcount  : integer ;

  BEGIN

    FOR bytecount := 1 to 8 DO
      BEGIN
        resultvalue := 0 ;
        k := 1 ;
        firstvalue  := store[stack + 8 + bytecount    ] ;
        secondvalue := store[stack + bytecount] ;
        FOR bitcount := 1 to 8 DO
          BEGIN
            IF ( odd(firstvalue) )  and ( odd(secondvalue) )
              THEN
                resultvalue := resultvalue + k ;
            k := k * 2 ;
            firstvalue  := firstvalue  div 2 ;
            secondvalue := secondvalue div 2
          END ;
        store[stack + 8 + bytecount] := resultvalue ;
      END;
    stack := stack + 8 ;

  END ; (* INT *)

PROCEDURE IOR   ;
(***************)

  (* Performs OR on two booleans on the stack,
     leaves boolean result on stack.
     Booleans placed on stack as two bytes,
     low byte contains 0 if false.
     evaluates expression :
     result := first boolean OR  second boolean.
     (true is 1 if first or second true)

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  boolean/ result    !
            +----------------------------+
            !  second boolean            ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  BEGIN

    IF (store[stack + 1] = 1) OR (store[stack + 3] = 1)
      THEN
        store[stack + 3] := 1
      ELSE
        store[stack + 3] := 0 ;
    stack := stack + 2 ;

  END ; (* IOR *)

PROCEDURE MMOD   ;
(***************)

  (* modulus of two integers on stack, leaves integer
     result on stack. Signed division.
     evaluates expression :
         result := first value mod second value

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  value / result     !
            +----------------------------+
            !  second value              ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  VAR

    result : integer ;

  BEGIN

    result := ( store[stack + 3] + 256 * store[stack + 4] ) mod
              ( store[stack + 1] + 256 * store[stack + 2] )  ;
    store[stack + 3] := result mod 256 ;
    store[stack + 4] := result div 256 ;
    stack := stack + 2

  END ; (* MMOD *)

PROCEDURE MPI   ;
(***************)

  (* multiplies two integers on stack, leaves integer
     result on stack. Signed division.
     evaluates expression :
         result := first value * second value

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  value / result     !
            +----------------------------+
            !  second value              ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  VAR

    result : integer ;

  BEGIN

    result := ( store[stack + 3] + 256 * store[stack + 4] ) *
              ( store[stack + 1] + 256 * store[stack + 2] )  ;
    store[stack + 3] := result mod 256 ;
    store[stack + 4] := result div 256 ;
    stack := stack + 2

  END ; (* MPI *)

PROCEDURE NGI   ;
(***************)

  (* changes sign of integer on stack
     evaluates expression :
         result := - result

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !     result                 !
            +----------------------------+
    stack ->!                            !<- stack  *)

  BEGIN

    store[stack + 1] := -store[stack + 1] ;
    store[stack + 2] := -store[stack + 2] ;

  END ;

PROCEDURE MNOT   ;
(***************)

  (* complements boolean value on stack
     evaluates expression :
         result := not result

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !     result                 !
            +----------------------------+
    stack ->!                            !<- stack  *)

  BEGIN

    IF store[stack + 1] <> 0
      THEN
        store[stack + 1] := 0
      ELSE
        store[stack + 1] := 1 ;

  END ; (* MNOT *)

PROCEDURE SBI   ;
(***************)

  (* subtracts two integers on stack, leaves integer
     result on stack.
     evaluates expression :
         result := first value - second value

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  first  value / result     !
            +----------------------------+
            !  second value              ! <- stack
            +----------------------------+
    stack ->!                            !           *)

  VAR

    result : integer ;

  BEGIN

    result := ( store[stack + 3] + 256 * store[stack + 4] ) -
              ( store[stack + 1] + 256 * store[stack + 2] )  ;
    store[stack + 3] := result mod 256 ;
    store[stack + 4] := result div 256 ;
    stack := stack + 2

  END ; (* SBI *)

PROCEDURE SGS   ;
(***************)

  (* Generate singleton set on stack.
     On stack number is present as integer
     of set member to be set, other set members
     not present.

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->!  setnumber / set           !<-- stack
            +----------------------------+
            !                            !           *)

  VAR

    count, setnumber,k : integer ;

  BEGIN

    setnumber := store[stack + 1] ;
    stack := stack - 6 ;
    FOR count := 1 TO 8 DO
      store [stack + count] := 0 ;
    count := setnumber div 8 ;
    setnumber := setnumber mod 8 ;
    k := 128 ;
    while setnumber > 0 do
      begin
        k := k div 2 ;
        setnumber := setnumber - 1
      end ;
    store[stack + 1 + count] := k ;

  END ;

PROCEDURE UNI   ;
(***************)

  (* gives the union of two sets on the stack,
     leaves result on the stack. Collects in result
     all elements present in either input set.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  first set / result set    !
            +----------------------------+
            !  second set                ! <- stack
            +----------------------------+
    stack ->!                            !            *)

VAR

  firstvalue, secondvalue, resultvalue,
  k, bytecount, bitcount  : integer ;

  BEGIN

    FOR bytecount := 1 to 8 DO
      BEGIN
        resultvalue := 0 ;
        k := 1 ;
        firstvalue  := store[stack + 8 + bytecount    ] ;
        secondvalue := store[stack + bytecount] ;
        FOR bitcount := 1 to 8 DO
          BEGIN
            IF ( odd(firstvalue) ) or ( odd(secondvalue) )
              THEN
                resultvalue := resultvalue + k ;
            k := k * 2 ;
            firstvalue  := firstvalue  div 2 ;
            secondvalue := secondvalue div 2
          END ;
        store[stack + 8 + bytecount] := resultvalue ;
      END;
    stack := stack + 8 ;

  END ; (* UNI *)

PROCEDURE LNC   ;
(***************)

  (* loads negative value on stack with value after opcode

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
    stack ->!  negative value            !
            +----------------------------+
            !                            !<- stack  *)

  BEGIN

    stack := stack - 2 ;
    store[stack + 1] := - store [pc    ] ;
    store[stack + 2] := - store [pc + 1] ;
    pc := pc + 2

  END ; (* LNC *)

PROCEDURE DEC   ;
(***************)

  (* decrement value on stack with value after opcode and
     leave result in same stack position

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  value /value decremented  !
            +----------------------------+
    stack ->!                            !<- stack  *)

  VAR

    value : integer ;

  BEGIN

    value := store[stack + 1] + 256 * store[stack + 2] -
             256 * store[pc] - store[pc + 1] ;
    store[stack + 1] := value mod 256 ;
    store[stack + 2] := value div 256 ;
    pc := pc + 2 ;

  END ;  (* DEC *)

PROCEDURE INC   ;
(***************)

  (* increment value on stack with value after opcode and
     leave result in same stack position

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  value /value incremented  !
            +----------------------------+
    stack ->!                            !<- stack  *)

  VAR

    value : integer ;

  BEGIN

    value := store[stack + 1] + 256 * store[stack + 2] +
             256 * store[pc] + store[pc + 1] ;
    store[stack + 1] := value mod 256 ;
    store[stack + 2] := value div 256 ;
    pc := pc + 2 ;

  END ;  (* INC *)

PROCEDURE ENT   ;
(***************)

  (* Reserve stack space for all variables local
     to procedure.
     PC (high), PC + 1 (low) contain nr of bytes to reserve.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
    stack ->!  room for local variables  !
                          .
                          .
            !                            !
            +----------------------------+
            !                            !<- stack  *)


  BEGIN

    stack := stack - ( store [pc] * 256 ) - store [pc + 1] ;
    pc := pc + 2

  END ;

PROCEDURE CAS   ;
(***************)

  (* processes case - statement.
     Following the opcode the complete jump-table
     is present including the otherwise part
     and the beginning and endvalue of the label:

     Stack layout
     begin                                 end
            !                            !<- stack
            +----------------------------+
    stack ->!  current label value       !
            +----------------------------+
            !                            !


   jumptable (from low to high addresses):

            +----------------------------+
            !  start value label in list !
            +----------------------------+
            !  end   value label in list !
            +----------------------------+
            !  address of otherwise part !    (if not used zero)
            +----------------------------+
            !  address of first label    !
            !   routine                  !
            +----------------------------+
            !                            !
                          .
         ordered next label routine addresses
                  (if not used zero)
                          .
            !                            !
            +----------------------------+
            !  address of last label     !
            !   routine                  !
            +----------------------------+    *)


  VAR

    startvalue , endvalue,
    curvalue, newpc         : integer ;

  BEGIN

    startvalue := store[pc + 1] + 256 * store[pc    ] ;
    endvalue   := store[pc + 3] + 256 * store[pc + 2] ;
    curvalue   := store[stack + 1] + 256 * store[stack + 2] ;
    stack := stack + 2 ;

    newpc := 0 ;

    (* current label value possibly in label list *)

    IF (curvalue >= startvalue) AND (curvalue <= endvalue)
      THEN
        BEGIN
          curvalue := pc + 6 + 2 * (curvalue - startvalue) ;
          newpc := store[curvalue + 1] + 256 * store[curvalue]
        END ;

    (* otherwise part or case-value not present in list *)

    IF newpc = 0
      THEN
        newpc := store[pc + 5] + 256 * store[pc + 4] ;

    (* no otherwise part *)

    IF newpc = 0
      THEN
        newpc := pc + 6 + ( (endvalue - startvalue + 1) * 2) ;
    pc := newpc ;

  END ; (* CASE *)

PROCEDURE MOVMVB  ;
(*****************)

  (* moves structures in memory
     Size to move follows opcode as word.
     If MVB then blanks to pad follows size as word.
     Source and destination address are on the stack.

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  destination address       ! <- stack
            +----------------------------+
            !  source  address           !
            +----------------------------+
    stack ->!                            !           *)

VAR

  source, dest, count, blanks, i : integer ;

  BEGIN

    source := store[stack + 1] + 256 * store[stack + 2] ;
    dest   := store[stack + 3] + 256 * store[stack + 4] ;
    count  := store[pc] + store[pc + 1] ;
    IF opcode = 196
      THEN
        blanks := store[pc + 2] + store[pc + 3] ;
    FOR i := 0 to (count - 1) do
      store[dest + i] := store[source + i] ;
    IF opcode = 196
      THEN
        FOR i := 0 to (blanks - 1) do
          store[dest + i + count] := ord(' ') ;
    stack := stack + 4 ;
    pc := pc + 2 ;
    IF opcode = 196
      THEN
        pc := pc + 2 ;

  END ; (* MOV *)

PROCEDURE DEC1  ;
(***************)

  (* decrement first value on stack with one and
     leave result in same stack position

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  value /value - 1          !
            +----------------------------+
    stack ->!                            !<- stack  *)

  VAR

    value : integer ;

  BEGIN

    value := store[stack + 1] + 256 * store[stack + 2] - 1 ;
    store[stack + 1] := value mod 256 ;
    store[stack + 2] := value div 256 ;

  END ;  (* DEC1 *)

PROCEDURE INC1  ;
(***************)

  (* increment first value on stack with one and
     leave result in same stack position

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  value /value + 1          !
            +----------------------------+
    stack ->!                            !<- stack  *)

  VAR

    value : integer ;

  BEGIN

    value := store[stack + 1] + 256 * store[stack + 2] + 1 ;
    store[stack + 1] := value mod 256 ;
    store[stack + 2] := value div 256 ;

  END ;  (* INC1 *)

PROCEDURE LDCS  ;
(***************)

  (* loads constant set following opcode on the
     stack.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
    stack ->!  set value                 !
            +----------------------------+
            !                            !<- stack  *)

  VAR

    i : integer ;

  BEGIN

    stack := stack - 8 ;
    FOR i := 0 to 7 do
      store [stack + i + 1] := store [pc + i] ;
    pc := pc + 8

  END ;

PROCEDURE CAP   ;
(***************)

{ The CAP instruction is used to call an assembly-language routine 
  previously loaded into a specified address. 
  The actual action of this instruction may be considered to be system-dependent. 
  Local parameters are stored at the stack with LDCI

     Stack layout
     begin                                 end
            !                            ! <- stack
            +----------------------------+
          ->!  External address          !
            +----------------------------+
    Stack ->!                            !
}

  BEGIN
    NotImplemented ;
    pc := pc + 2
  END ;

PROCEDURE LCA   ;
(***************)

  (* Load address of strng following opcode on stack
     pc incremented up to opcode following strng

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
    stack ->!  address of strng        !
            +----------------------------+
            !                            !<- stack  *)

  BEGIN

    store [stack     ] := (pc + 1) div 256 ;
    store [stack - 1 ] := (pc + 1) mod 256 ;
    stack := stack - 2 ;
    pc := pc + store[pc] + 1 ;

  END ;

PROCEDURE CSP   ;
(***************)

  (* standard procedures *)

  PROCEDURE WRI ;
  (*************)

  (* writes integer on stack in specified field width.

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  char code                 !<- stack
            +----------------------------+
            ! field width                !
            +----------------------------+
    stack ->!                            !
                                              *)

    VAR

      width, value, i, j : integer ;
      sign : boolean ;
      intstr : array [1 .. 6] of char ;

    BEGIN

      width := store[stack + 2] * 256 + store[stack + 1] ;
      value := store[stack + 4] * 256 + store[stack + 3] ;
      sign := ( value < 0 ) ;
      value := abs (value) ;
      i := 6 ;
      REPEAT
        intstr[i] := chr ( (value mod 10) + ord ('0') ) ;
        value := value div 10 ;
        i := i - 1
      UNTIL value = 0 ;
      IF sign
        THEN
          BEGIN
            intstr[i] := '-' ;
            i := i - 1
          END ;
      FOR j := 1 TO i DO
        intstr[j] := ' ' ;
      IF width < 6 - i
        THEN
          width := 6 - i ;
      FOR i := (6 - width + 1) TO 6 DO
        PutChar ( intstr[i] ) ;
      stack := stack + 4

    END ; (* WRI *)

  PROCEDURE WRC ;
  (*************)

  (* writes character on stack in specified field width.
     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  char code                 ! <- stack
            +----------------------------+
            ! field width                !
            +----------------------------+
    stack ->!                            !
                                              *)
    VAR

      width, i : integer ;

    BEGIN

      width := store[stack + 2] * 256 + store[stack + 1] ;
      IF width > 1
        THEN
          FOR i := 1 TO (WIDTH - 1) DO
            PutChar ( ' ' ) ;
      PutChar ( chr ( store[stack + 3] ) ) ;
      stack := stack + 4

    END ; (* WRC *)

  PROCEDURE WRS ;
  (*************)

    (* prints strngto output,

     Stack layout
     begin                                 end
            !                            !
            +----------------------------+
            !  address                   !<- stack
            !  of strng                !
            +----------------------------+
            ! specified                  !
            ! field width                !
            +----------------------------+
            ! actual                     !
            ! field width                !
            +----------------------------+
    stack ->!                            !
                                              *)

    VAR

      actwidth, spcwidth, i, addr : integer ;

    BEGIN

      actwidth := store[stack + 2] * 256 + store[stack + 1] ;
      spcwidth := store[stack + 4] * 256 + store[stack + 3] ;

      IF spcwidth > 0
        THEN
          BEGIN
            addr := (spcwidth - actwidth) ;
            IF addr > 0
              THEN
                FOR i := 0 TO addr DO
                  PutChar ( ' ' ) ;

            addr := store[stack + 6] * 256 + store[stack + 5] ;
            FOR i := 0 to (spcwidth - 1) DO
            PutChar ( chr ( store[addr + i]) ) ;
          END ;

      stack := stack + 6 ;

    END ;  (* WRS *)

  PROCEDURE RDI ;
  (*************)


    (* reads integer
       if terminal then
         if CR entered then enofline becomes true
            and a space returned
         if CTRL-Z entered then endoffile and endofline
            become true
       at least the value zero is returned,
       also for illegal characters

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  address for integer       !<- stack
            +----------------------------+
    stack ->!                            !        *)

    VAR

      addr,value : integer ;
      skipblank : boolean ;

    BEGIN

      addr := store[stack + 1] + 256 * store[stack + 2] ;
      value := 0 ;
      skipblank := true ;
      WITH pfiles[filenumber]^ DO
        BEGIN
         IF not endofline
           THEN
             REPEAT
               kar := GetChar  ;
               IF filenumber <> 1
                 THEN
                   BEGIN
                     IF endofline
                       THEN
                         kar := chr(0)
                   END
                 ELSE
                   BEGIN
                     IF kar = chr(CR)
                       THEN
                         BEGIN
                           kar := chr(0) ;
                           endofline := true
                         END
                     ELSE IF kar = chr(CTRLZ)
                       THEN
                         BEGIN
                           kar := chr(0) ;
                           endofline := true ;
                           endoffile := true
                         END
                     ELSE
                       BEGIN
                          endoffile := true ;
                         endofline := true ;
                       END
                   END ;
               IF ( ( ord(kar) >= ord ('0') ) AND
                    ( ord(kar) <= ord ('9') )     )
                 THEN
                   BEGIN
                     value := 10 * value + ord (kar) - ord('0') ;
                     skipblank := false ;
                     IF value > 32767
                       THEN
                         BEGIN
                           value := 32767 ;
                           kar := chr(0) ;
                         END
                   END
                 ELSE IF (kar = ' ') AND ( NOT (skipblank) )
                   THEN
                     kar := chr(0)
                 ELSE IF NOT (kar = ' ')
                   THEN
                     kar := chr(0)
             UNTIL ( kar = chr(0) )

        END ;

      store[addr    ] := value mod 256 ;
      store[addr + 1] := value div 256 ;

      stack := stack + 2 ;

    END ; (* RDI *)

  PROCEDURE RLN ;
  (*************)

  VAR

    kar : char ;

  (* if already endofline then endofline is reset
     else characters read from input until endofline passed *)

    BEGIN

      WITH pfiles[filenumber]^ DO
        BEGIN
          IF filenumber = 1
            THEN
              BEGIN
                IF endofline
                  THEN
                      endofline := false
                  ELSE
                    REPEAT
                      kar := GetInputChar
                    UNTIL kar = chr(CR) ;
              END
            ELSE
              GetLineFile(pfiles[filenumber])
        END

    END ; (* RLN *)

  PROCEDURE RDC ;
  (*************)

    (* reads character from input file

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  address for char          !<- stack
            +----------------------------+
    stack ->!                            !        *)


    VAR

      addr : integer ;
      kar  : char ;

    BEGIN

      addr := store[stack + 1] + 256 * store[stack + 2] ;
      WITH pfiles[filenumber]^ DO
        BEGIN
          IF filenumber = 1
            THEN
              BEGIN
                kar := GetInputChar  ;
                IF kar = chr(CR)
                  THEN
                    BEGIN
                      endofline := true ;
                      kar := ' '
                    END
                ELSE IF kar = chr(CTRLZ) THEN
                  BEGIN
                    endoffile := true ;
                    endofline := true ;
                    kar := ' '
                  END
                ELSE
                  BEGIN
                    endoffile := false ;
                    endofline := false  ;
                  END
              END
            ELSE
              kar := GetChar
          END ;

      store[addr] := ord (kar) ;
      stack := stack + 2 ;

    END ;

  PROCEDURE WLN ;
  (*************)

    (* prints Carriage return and Linefeed on output
       else writes end of line to file               *)

    BEGIN

      IF filenumber = 2
        THEN
          BEGIN
            PutChar ( chr(CR) ) ;
            PutChar ( chr(LF) )
          END
        ELSE
          PutLineFile(pfiles[filenumber]) ;

    END ; (* WLN *)

  PROCEDURE NWW ;
  (*************)

     (* Reserves space in the heap for variable
        pointer address and size to reserve on the stack
        heap pointer updated

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  address of pointer        !<- stack
            +----------------------------+
            !  size to reserve on heap   !
            +----------------------------+
    stack ->!                            !        *)


    VAR

      addr : integer ;

    BEGIN

      addr := store [stack + 3] + 256 * store[stack + 4] ;
      store [addr    ] := heap mod 256 ;
      store [addr + 1] := heap div 256 ;
      heap := heap + store[stack + 1] + 256 * store[stack +2] ;
      stack := stack + 4 ;

    END ;

  PROCEDURE EFF ;
  (*************)

     (* enters boolean end of file on stack

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->!  boolean (1 if endoffile)  !
            +----------------------------+
            !                            !<- stack        *)

    BEGIN

      stack := stack - 2 ;
      WITH pfiles[filenumber]^ DO
        BEGIN
          IF InOut <> INP
            THEN
              FatalFileError(7, pfiles[filenumber]) ;
          IF filestatus <> opened
            THEN
              FatalFileError(5, pfiles[filenumber]) ;

          IF endoffile
            THEN
              store[stack + 1] := 1
            ELSE
              store[stack + 1] := 0 ;
          store[stack + 2] := 0
        END

    END ; (* EFF *)

  PROCEDURE RST ;
  (*************)

     (* Restores heap pointer to value in pointer
        address of pointer on the stack

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            !  address of pointer        !<- stack
            +----------------------------+
    stack ->!                            !        *)

    VAR

      addr : integer ;

    BEGIN

      addr := store [stack + 1] + 256 * store[stack + 2] ;
      heap := store[addr ] + 256 * store[addr + 1] ;
      stack := stack + 2 ;

    END ;

  PROCEDURE ELN ;
  (*************)

     (* enters boolean end of line on stack

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->!  boolean (1 if endofline)  !
            +----------------------------+
            !                            !<- stack        *)

    BEGIN

      stack := stack - 2 ;
      WITH pfiles[filenumber]^ DO
        BEGIN
          IF InOut <> INP
            THEN
              FatalFileError(7, pfiles[filenumber]) ;
          IF filestatus <> opened
            THEN
              FatalFileError(5, pfiles[filenumber]) ;
          IF endofline
            THEN
              store[stack + 1] := 1
            ELSE
              store[stack + 1] := 0 ;
          store[stack + 2] := 0 ;
        END

    END ; (* ELN *)

  PROCEDURE STP ;
  (*************)

  (* Stops execution of interpreter immediately *)

    BEGIN

      continue := false ;
      if debug
        then
          writeln('Pascal-M stopped') ;

    END ;

  PROCEDURE ODDM ;
  (**************)

     (* enters boolean true on stack if value on stack is odd

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->!  value/ result (1 if odd)  !<-stack
            +----------------------------+
            !                            !        *)

    BEGIN

      IF ODD ( store[stack + 1] )
        THEN
          store[stack + 1] := 1
        ELSE
          store[stack + 1] := 0 ;
      store[stack + 2] := 0 ;

    END ; (* ODDM *)


  PROCEDURE STT ;
  (*************)

     (* enters status of file on stack for current filenumber

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->!  value/ result (1 if odd)  !
            +----------------------------+
            !                            !<-stack        *)

    BEGIN

      stack := stack - 2 ;
      store[stack + 2] := 0 ;
      WITH pfiles[filenumber]^ DO
        BEGIN
          IF filestatus = opened
            THEN
              store[stack + 1] := 0
          ELSE IF filestatus = fatalerror
            THEN
              store[stack + 1] := -1
          ELSE IF filestatus = closed
            THEN
              store[stack + 1] := 1 ;
        END

    END ; (* STT *)

  PROCEDURE RSFRWW ;
  (****************)

    (* resets or rewrites file
       filename as string on stack


     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            ! address of strng         !<-- stack
            +----------------------------+
            ! length of strng          !
            +----------------------------+
    stack ->!                            !        *)

    VAR

      i, len, addr  : integer ;

    BEGIN

      len  := store[stack + 1] + 256 * store[stack + 2] ;
      addr := store[stack + 3] + 256 * store[stack + 4] ;
      WITH pfiles[filenumber]^ DO
        BEGIN
          IF store[pc] = 13
            THEN
              InOut := INP
            ELSE
              InOut := OUTP ;
          (* copy filename , address on stack to
             fileinfo, pad with blanks *)
          FOR i := 1 TO 20 DO
            name[i] := ' ' ;
          IF len > 20
            THEN
              len := 20 ;
          FOR i := 1 TO len DO
            name[i] := chr(store[addr + i - 1]) ;
          openfile(pfiles[filenumber])
        END ;

      stack := stack + 4 ;

    END ; (* RSFRWW *)

    PROCEDURE CLS ;
    (*************)

    (* closes file pointed to by current filenumber *)

    BEGIN

      closefile(pfiles[filenumber]) ;

    END ; (* CLS *)


    BEGIN (* main of Call Standard Procedure *)

      CASE store[pc] OF
       0 : WRI  ;
       1 : WRC  ;
       2 : WRS  ;
       3 : RDI  ;
       4 : RLN  ;
       5 : RDC  ;
       6 : WLN  ;
       7 : NWW  ;
       8 : EFF  ;
       9 : RST  ;
      10 : ELN  ;
      11 : STP  ;
      12 : ODDM ;
      13 : RSFRWW  ;
      14 : RSFRWW  ;
      15 : STT  ;
      16 : CLS  ;
     END ;
    pc := pc + 1

  END ; (* CSP *)

PROCEDURE CUP1  ;
(***************)

  (* call user procedure , number of procedure follows opcode
     save return PC of instruction following CUP1 in frame

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            ! base address link pointer  !
            +----------------------------+
            ! saved markpointer          !
            +----------------------------+
            ! saved pc                   !
            +----------------------------+
   stack -> !                            ! <- stack        *)

  BEGIN

    markp := markpsave ;
    store [ markp - 4] := (pc + 1) div 256 ;
    store [ markp - 5] := (pc + 1) mod 256 ;
    pc := procbuffer[ store[ pc ] ].procaddress ;

  END ;

PROCEDURE CUP2  ;
(***************)

  BEGIN
    NotImplemented ;
    pc := pc + 1
  END ;

PROCEDURE FIX21 ;
(***************)

  (* Transfers one byte item on stack into two byte item,
     byte item moved to low byte of two byte item,
     high byte filled with zero

     Stack layout :
     start                                  end
            !  byte item / 0             !
            +----------------------------+
   stack -> !          ? / byte item     !
            +----------------------------+
            !                            !<- stack        *)

  BEGIN

    store[stack] := store[stack + 1] ;
    store[stack + 1] := 0 ;
    stack := stack - 1 ;

  END ; (* FIX21 *)

PROCEDURE LNS   ;
(***************)

  (* Load null set on stack : 8 bytes of zero

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->!                            !
                  8 bytes with  zero
            !                            !
            +----------------------------+
            !                            ! <- stack  *)

  VAR

    i : integer ;

  BEGIN

    stack := stack - 8 ;
    FOR i := 1 TO 8 DO
      store[stack + i] := 0

  END ; (* LNS *)

PROCEDURE SFA   ;
(***************)

  (* Set file address for all file operations to follow
     filenumber in least sign byte on stack

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
            ! filenumber                 !  <- stack
            +----------------------------+
    stack ->!                            !  *)


  BEGIN

    filenumber := store[stack + 1] ;
    stack := stack + 2 ;

  END ; (* SFA *)

PROCEDURE GFA   ;
(***************)

  (* Get file address for all file operations to follow
     filenumber in least sign byte on stack

     Stack layout :
     start                                  end
            !                            !
            +----------------------------+
    stack ->! filenumber                 !
            +----------------------------+
            !                            ! <- stack  *)


  BEGIN

    store[stack ] :=  filenumber  ;
    store[stack +1] := 0 ;
    stack := stack - 2 ;

  END ; (* SFA *)



(************************************************************)
(*							    *)
(*    END OF INSTRUCTION INTERPRETATION PART 		    *)
(*							    *)
(************************************************************)


(************************************************************)
(*							    *)
(*    DEBUG part					    *)
(*    Hereafter follow the procedures for debugging         *)
(*    the M-code instructions                               *)
(*							    *)
(************************************************************)


FUNCTION GetHex : integer ;
(*************************)

  VAR

     i : integer ;
     strldaddr : alpha ;

  BEGIN

    FOR i := 1 TO alphalength DO
      strldaddr[i] := ' ' ;
    i := 1 ;
    REPEAT
      read ( strldaddr[i] ) ;
      i := i + 1
    UNTIL ( eoln ) or (i > alphalength ) ;
    readln ;
    GetHex := ReadHex ( strldaddr, i ) ;

  END ; (* GetHex *)

PROCEDURE BreakCommand ;
(**********************)

  (* Handles debug break subcommand *)

  VAR

    comkar : char ;
    i,j : integer ;

  BEGIN

     write ('Break subcommand > ') ;
     readln ( comkar ) ;

     IF comkar = '?'
       THEN
         BEGIN
           writeln ;
           writeln ( 'Available Break commands :') ;
           writeln ;
           writeln ( '  C = Clear all breakpoints');
           writeln ( '  D = Show  all breakpoints');
           writeln ( '  S = Set breakpoint');
           writeln ;
           BreakCommand
         END
      ELSE IF (comkar = 'C') OR (comkar = 'c')
        THEN
          BEGIN
            FOR i := 1 to maxbreak DO
              breaktable[i] := 32767 ;
            break := false
          END
      ELSE IF (comkar = 'D') OR (comkar = 'd')
        THEN
          BEGIN
            writeln ;
            FOR i := 1 to maxbreak DO
              BEGIN
                write('Breakpoint ', i:1, ' = ') ;
                IF breaktable[i] <> 32767
                  THEN
                    BEGIN
                      writehex (breaktable[i] div 256) ;
                      writehex (breaktable[i] mod 256)
                    END
                  ELSE
                      write  ('clear') ;
                writeln
              END
          END
      ELSE IF (comkar = 'S') OR (comkar = 's')
        THEN
          BEGIN
            write ('Address of breakpoint = ') ;
            i := GetHex ;
            REPEAT
              write ('Breakpoint number = ' ) ;
              j := GetHex
            UNTIL (j > 0) AND (j < maxbreak + 1) ;
            breaktable[j]  := i ;
            break := true
          END

  END ; (* BreakCommand *)

PROCEDURE CommandDebug ;
(**********************)

  (* gives debug information *)

  VAR

    comkar : char ;
    dispstart, dispend, i,j : integer ;

  BEGIN

    REPEAT
     writeln ;
     write ( 'PC    = ') ;
     writehex ( PC div 256 ) ;
     writehex ( PC mod 256 ) ;
     writeln ;
     write ('M-Debug > ') ;
     readln ( comkar ) ;

     IF comkar = '?'
       THEN
         BEGIN
           writeln ;
           writeln ( 'Available debug commands :') ;
           writeln ;
           writeln ( '  B = Breakpoint commands') ;
           writeln ( '  C = Show stack contents' );
           writeln ( '  D = Display store') ;
           writeln ( '  E = Show stack contents') ;
           writeln ( '  G = Continue without debugger') ;
           writeln ( '  H = Show heappointer') ;
           writeln ( '  I = Insert in store') ;
           writeln ( '  M = Show markpointer') ;
           writeln ( '  N = Next instruction executed') ;
           writeln ( '  P = Show programcounter') ;
           writeln ( '  S = Show stackpointer') ;
           writeln ( '  T = Show procedure table' ) ;
           writeln ( '  X = Exit interpreter') ;
         END
     ELSE if (comkar = 'B') or (comkar = 'b')
       THEN
         BreakCommand
     ELSE if (comkar = 'I') or (comkar = 'i')
       THEN
         BEGIN
           write ('Address = ' ) ;
           dispstart := GetHex ;
           write ('Value = ' ) ;
           dispend := GetHex ;
           IF (dispstart > 0) and (dispstart < maxstore)
             THEN
               store [ dispstart ] := dispend
             ELSE
               writeln ('Address out of range' )
         END
     ELSE if (comkar = 'D') or (comkar = 'd')
       THEN
         BEGIN
           write ('Start = ') ;
           dispstart := GetHex ;
           write ('End   = ') ;
           dispend := GetHex ;
           FOR i := dispstart to dispend do
             BEGIN
               writehex ( i div 256 ) ;
               writehex ( i mod 256 ) ;
               write ( ' = ' ) ;
               writehex ( store [ i ] ) ;
               IF i = stack
                 THEN
                   writeln ('   <---- stackpointer')
               ELSE IF i = markp
                 THEN
                   writeln ('   <---- markpointer')
               ELSE IF i = markpsave
                 THEN
                   writeln ('   <---- marksave pointer')
               ELSE IF i = heap
                 THEN
                   writeln ('   <---- heappointer')
               ELSE
                 writeln ;
             END
         END
     ELSE if (comkar = 'P') or (comkar = 'p')
       THEN
         BEGIN
           write ( 'PC    = ') ;
           writehex ( PC div 256 ) ;
           writehex ( PC mod 256 ) ;
         END
     ELSE if (comkar = 'S') or (comkar = 's')
       THEN
         BEGIN
           write ( 'Stack = ') ;
           writehex ( stack div 256 ) ;
           writehex ( stack mod 256 ) ;
         END
     ELSE if (comkar = 'C') or (comkar = 'c')
       THEN
         BEGIN
           writeln ('Stack contents :') ;
           i := stack - 6 ;
           WHILE ( i <= (stack + 14) ) and ( i < 32767 ) DO
             BEGIN
               writehex ( i div 256 ) ;
               writehex ( i mod 256 ) ;
               write ( ' = ' ) ;
               writehex ( store [ i ] ) ;
               IF i = stack
                 THEN
                   writeln ('   <---- stackpointer')
               ELSE IF i = markp
                 THEN
                   writeln ('   <---- markpointer')
               ELSE IF i = markpsave
                 THEN
                   writeln ('   <---- marksave pointer')
               ELSE IF i = heap
                 THEN
                   writeln ('   <---- heappointer')
               ELSE
                 writeln ;
               i := i + 1
             END
         END
     ELSE if (comkar = 'H') or (comkar = 'h')
       THEN
         BEGIN
           write ( 'Heap  = ') ;
           writehex ( heap div 256 ) ;
           writehex ( heap mod 256 ) ;
         END
     ELSE if (comkar = 'M') or (comkar = 'm')
       THEN
         BEGIN
           write ( 'Markp = ') ;
           writehex ( markp div 256 ) ;
           writehex ( markp mod 256 ) ;
         END
     ELSE if (comkar = 'X') or (comkar = 'x')
       THEN
         BEGIN
           continue := false ;
           comkar := 'X'
         END
     ELSE if (comkar = 'G') or (comkar = 'g')
       THEN
         BEGIN
           debug := false ;
           comkar := 'G'
         END
     ELSE if (comkar = 'T') or (comkar = 't')
       THEN
         FOR i := 0 TO (procf - 1 ) DO
           BEGIN
             writehex ( procaddr div 256) ;
             writehex ( procaddr mod 256) ;
             write ( ': ' ) ;
             writehex( procbuffer[i].procaddress mod 256 ) ;
             write ( ' ' ) ;
             writehex( procbuffer[i].procaddress div 256 ) ;
             write ( ' , number = ', i:2, ' , name = ' ) ;
             for j := 1 to shlen do
             write ( procbuffer[i].procname[j] ) ;
             writeln ;
             procaddr := procaddr + 2
           END ;

    UNTIL (comkar = 'X' ) or
          (comkar = 'n')  or
          (comkar = 'N')  or
          (comkar = 'G') ;

  END ; (* Debug *)


(************************************************************)
(*    Interpreter					    *)
(*    Hereafter follows the main loop procedure of the      *)
(*    interpreter for the M-code instructions               *)
(*							    *)
(************************************************************)

PROCEDURE Interpr ;
(*****************)

  VAR

    i : integer ;

  BEGIN

    FOR i := 1 to maxbreak DO
      breaktable[i] := 32767 ;
    break := false ;
    WHILE continue DO
      BEGIN
        IF debug
          THEN
            CommandDebug ;
        IF break
          THEN
            BEGIN
              i := 0 ;
              REPEAT
                i := i + 1
              UNTIL (i = (maxbreak + 1)) OR (breaktable[i] = pc) ;
              IF i < maxbreak + 1
                THEN
                  BEGIN
                    writeln ('Breakpoint ', i:2) ;
                    debug := true ;
                    CommandDebug
                  END
            END ;
        opcode := store[pc] ;
        pc := pc + 1 ;
        IF opcode < 144
          THEN
            BEGIN
              opcode := store[pc - 1] div 16 ;
              level  := store[pc - 1] mod 16 ;
              CASE opcode of
                0 : LDCIS ;
                1 : LDAS  ;
                2 : LDA   ;
                3 : MST0  ;
                4 : MSTN  ;
                5 : LOD1  ;
                6 : LOD2  ;
                7 : STR1  ;
                8 : STR2
              END
            END
          ELSE
            CASE opcode of
              144 :   LEQ2  ;
              145 :   MFOR   ;
              146 :   LEQM  ;
              147 :   LES2  ;
              148 :   LEQ8  ;
              149 :   LESM  ;
              150 :   EQU2  ;
              151 :   GEQ8  ;
              152 :   EQUM  ;
              153 :   EQU8  ;
              154 :   IND1  ;
              155 :   IND2  ;
              156 :   IND8  ;
              157 :   STO1  ;
              158 :   ST02  ;
              159 :   STO8  ;
              160 :   LDC   ;
              161 :   RETP  ;
              162 :   ADI   ;
              163 :   MAND  ;
              164 :   DIF   ;
              165 :   DVI   ;
              166 :   INN   ;
              167 :   INT   ;
              168 :   IOR   ;
              169 :   MMOD  ;
              170 :   MPI   ;
              171 :   NGI   ;
              172 :   MNOT  ;
              173 :   SBI   ;
              174 :   SGS   ;
              175 :   UNI   ;
              176 :   LNC   ;
              177 :   FJP   ;
              178 :   UJP   ;
              179 :   DEC   ;
              180 :   INC   ;
              181 :   ENT   ;
              182 :   CAS   ;
              183 :   MOVMVB ;
              184 :   DEC1  ;
              185 :   INC1  ;
              186 :   LDCS  ;
              187 :   CAP   ;
              188 :   LCA   ;
              189 :   CSP   ;
              190 :   CUP1  ;
              191 :   CUP2  ;
              192 :   FIX21 ;
              193 :   LNS   ;
              194 :   SFA   ;
              195 :   GFA   ;
              196 :   MOVMVB ;
              OTHERWISE
                BEGIN
                  write ('Unknown opcode at ') ;
                  WriteHex ( pc div 256 ) ;
                  WriteHex ( pc mod 256 ) ;
                  writeln ;
                  debug := true ;
                END

            END ; (* case *)

       IF  ( stack - heap ) < stackmin
          THEN
            BEGIN
              writeln ;
              write ('Stack overflow at ' ) ;
              WriteHex ( pc div 256 ) ;
              WriteHex ( pc mod 256 ) ;
              writeln ;
              halt ;
            END ;

      END ; (* while *)

  END ; (* Interpr *)


PROCEDURE InitializeStore ;
(*************************)

  var

    i : integer ;

  BEGIN

    for i := 1 to maxstore do
      store[i] := 0 ;


    (* Standard M-code procedure to start M-code interpretation,
       placed at end of available core                    *)

    store [maxstore - 4] := 48  ; (* MST0      *)
    store [maxstore - 3] := 190 ; (* CUP1      *)
    store [maxstore - 2] := 0   ; (* proc nr 0 *)
    store [maxstore - 1] := 189 ; (* CSP       *)
    store [maxstore    ] := 11  ; (* Standard proc Stop *)

  END ; (* InitializeStore *)
PROCEDURE OpenFiles ;
(*******************)

VAR

  filename : string ;


BEGIN (* Openfiles *)

    if paramcount > 0
      then
        filename := paramstr(1)
      else
        begin
          write('Object filename = ' ) ;
          readln(filename) ;
         end ;
    if not FileExists(filename)
      then
        begin
          writeln('No such file') ;
          halt(1) ;
        end
      else
        begin
          assign ( objectfile, filename);
          reset(objectfile) ;
         end ;

     debug := (paramstr(2) = 'debug') ;


END ; (* OpenFiles *)

(************************************************************)
(*							    *)
(*    Main of interpreter 				    *)
(*							    *)
(************************************************************)

  BEGIN

    (* initialize *)

    OpenFiles ;
    continue := true ;

    procf := 0 ;

    InitializeStore ;

    pc := 0 ;
    ldaddr := 0 ;
    loaderror := false ;
    Load ;
    IF loaderror
      then
        halt(1) ;
    (* leave cursor on first position new line *)
    writeln ;
    writeln ;

    (* initialize all m-code machine registers *)
    heap := pc ;
    pc := maxstore - 4 ;
    stack := maxstore - 5 ;
    markp := stack ;
    FileInitialize ;
    Interpr ;

END.
